import copy
import datetime
import base64
import logging
import io
import os.path

log = logging.getLogger(__name__)
log.setLevel("DEBUG")

from intelmqmail.tableformat import build_table_formats, ExtraColumn, IntelMQColumn, format_as_csv
from intelmqmail.notification import Postponed
from intelmqmail.templates import full_template_filename

standard_column_titles = {
    # column titles for standard event attributes
    'classification.taxonomy' : 'classification_taxonomy',
    'classification.identifier' : 'tag',
    #'classification.identifier' : 'service', #drone_brute_force#
    'classification.type' : 'classification_type',
    'destination.account' : 'username',
    'destination.asn' : 'dst_asn', #sinkhole_http_drone#ipv6_sinkhole_http_drone#microsoft_sinkhole#outdated_dnssec_key#darknet#http_scanners#ics_scanners#
    #'destination.asn' : 'dest_asn', #drone_brute_force#
    #'destination.asn' : 'cc_asn', #drone#drone_spam#
    'destination.fqdn' : 'http_host', #sinkhole_http_drone#ipv6_sinkhole_http_drone#microsoft_sinkhole#
    #'destination.fqdn' : 'cc_dns', #drone#drone_spam#
    #'destination.fqdn' : 'dst_dns', #http_scanners#ics_scanners#
    #'destination.fqdn' : 'dest_dns', #drone_brute_force#
    'destination.geolocation.cc' : 'dst_geo', #sinkhole_http_drone#
    #'destination.geolocation.cc' : 'dest_geo', #drone_brute_force#
    #'destination.geolocation.cc' : 'cc_geo', #drone#drone_spam#
    'destination.geolocation.region' : 'dst_region', #ipv6_sinkhole_http_drone#
    'destination.ip' : 'dst_ip',
    #'destination.ip' : 'dest_ip', #drone_brute_force#
    #'destination.ip' : 'cc_ip', #drone#drone_spam#
    'destination.port' : 'dst_port',
    #'destination.port' : 'dest_port', #drone_brute_force#
    #'destination.port' : 'cc_port', #drone#drone_spam#
    'destination.url' : 'url',
    #'destination.url' : 'http_url', #ipv6_sinkhole_http_drone#
    'event_description.text' : 'category',
    'malware.hash' : 'md5hash',
    'malware.name' : 'type', #sinkhole_http_drone#microsoft_sinkhole
    #'malware.name': 'infection', #drone#
    #'malware.name' : 'tag', #ipv6_sinkhole_http_drone#compromised_website
    'protocol.application' : 'application', #drone#drone_spam#compromised_website#
    #'protocol.application' : 'service', #drone_brute_force#
    #'protocol.application' : 'protocol', #ics_scanners#
    'protocol.transport' : 'protocol',
    'source.account' : 'username',
    'source.asn' : 'asn',
    'source.fqdn' : 'hostpname',
    'source.fqdn' : 'host',
    'source.fqdn' : 'http_host',
    #'source.asn' : 'src_asn', #ipv6_sinkhole_http_drone#
    'source.geolocation.cc' : 'geo',
    #source.geolocation.cc' : 'src_geo', #ipv6_sinkhole_http_drone#
    'source.geolocation.city' : 'city', #ALLA#
    'source.geolocation.region' : 'region',
    #source.geolocation.region' : 'src_region', #ipv6_sinkhole_http_drone#
    'source.ip' : 'ip', # Nästan alla#
    #'source.ip' : 'src', #spam_url"
    #'source.ip' : 'src_ip', #ipv6_sinkhole_http_drone#
    'source.local_hostname' : 'server_name', #open_mssql#
    #'source.local_hostname' : 'server_name', #open_ldap#
    'source.port' : 'port',
    #'source.port' : 'src_port', #sinkhole_http_drone#ipv6_sinkhole_http_drone#microsoft_sinkhole#amplification_ddos_victim
    'source.reverse_dns' : 'hostname',
    #'source.reverse_dns' : 'host', #spam_url#
    'source.tor_node' : 'tor', #sinkhole_http_drone#microsoft_sinkhole#
    'source.url' : 'url',
    'time.source' : 'timestamp', #ALLA#
    'user_agent' : 'agent', #2st
    #'user_agent' : 'http_agent', #sinkhole_http_drone#
    #'user_agent' : 'user_agent', #sandbox_url#

}

def add_default_titles(columns):
    """
    Add the standard title to each of the columns.

    Args:
        columns:

    Returns:

    """
    log.debug("add_default_titles; %r",columns)
    extended_columns = []
    for col in columns:
        if isinstance(col, str):
            if col.startswith("extra:"):
                extended_columns.append(ExtraColumn(col[6:],col[6:]))
            else:
                extended_columns.append(IntelMQColumn(standard_column_titles[col],col))
        else:
            extended_columns.append(col)
    log.debug("add_default_titles; %r",extended_columns)
    return extended_columns


# This function takes a list of tuples as argument,
# The tuples are made up of (<event_data_format>,[list of columns]
def table_formats_with_default_titles(formats):
    """
    Frontend for build_table_formats that adds standard column titles.

    Args:
        formats: a list of formats

    Returns:
        A TableFormat

    Each format is a tuple made up of a "key" that should match an event_data_format and a list of columns
    A column can be given in four different ways
        1) A single key string. The string will be translated using the standard_column_titles dictionary. 
           If it starts with extra: it will become an ExtraColumn instance, otherwise an IntelMQColumn instance
        2) A tuple("<key>","<title>"). The key must not start with extra:
        3) An IntelMQColumn("<title>", "<key>") 
        4) An ExtraColumn("<title>","<key>"). Please note that there is no "extra:" in the key
    Example; 
      table_formats_with_default_titles([
        (csv_Accessible-RDP", ["source.ip", 
                               "extra:handshake",
                               ("time.source", "timestamp"), 
                               IntelMQColumn("port_number", "source.port"),
                               ExtraColumn("certificate_length", "cert_length"),
                            ]),
        ])
    """
    log.debug("table_formats_with_default_titles: %r",formats)
    for name, columns in formats:
        log.debug("table_formats_with_default_titles: %r, %r",name, add_default_titles(columns))
    btf = build_table_formats([(name, add_default_titles(columns))
                                for name, columns in formats])

    log.debug("table_formats_with_default_titles: %r", btf)
    return btf
    return build_table_formats([(name, add_default_titles(columns))
                                for name, columns in formats])


# Used to create ${events_as_csv}
csv_formats = table_formats_with_default_titles([
    ('csv_Amplification-DDoS-Victim', 
        [
            'time.source',
            'source.ip',
            'protocol.transport',
            'destination.port',
            'extra:tag',
            'source.port',
            ('source.port', 'src_port'),
            'source.reverse_dns',
            'source.asn',
            'source.geolocation.cc',
            'source.geolocation.region',
            'source.geolocation.city',
            'extra:naics',
            'extra:sic', 
            'extra:request', 
            'extra:count',
            'extra:bytes',
            'extra:sensor_geo',
            'extra:sector',
            'extra:end_time',
            'extra:public_source',
    ]),
    ("csv_Accessible-FTP", [
        #"timestamp","ip","protocol","port","hostname","tag","asn","geo","region","city","naics","sic","banner","handshake","cipher_suite","cert_length","subject_common_name","issuer_common_name","cert_issue_date","cert_expiration_date","sha1_fingerprint","cert_serial_number","ssl_version","signature_algorithm","key_algorithm","subject_organization_name","subject_organization_unit_name","subject_country","subject_state_or_province_name","subject_locality_name","subject_street_address","subject_postal_code","subject_surname","subject_given_name","subject_email_address","subject_business_category","subject_serial_number","issuer_organization_name","issuer_organization_unit_name","issuer_country","issuer_state_or_province_name","issuer_locality_name","issuer_street_address","issuer_postal_code","issuer_surname","issuer_given_name","issuer_email_address","issuer_business_category","issuer_serial_number","sha256_fingerprint","sha512_fingerprint","md5_fingerprint","cert_valid","self_signed","cert_expired","validation_level","auth_tls_response","auth_ssl_response","tlsv13_support","tlsv13_cipher"
        'time.source',
        'source.ip',
        'protocol.transport',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:banner',
        'extra:handshake',
        'extra:cipher_suite',
        'extra:cert_length',
        'extra:subject_common_name',
        'extra:issuer_common_name',
        'extra:cert_issue_date',
        'extra:cert_expiration_date',
        'extra:sha1_fingerprint',
        'extra:cert_serial_number',
        'extra:ssl_version',
        'extra:signature_algorithm',
        'extra:key_algorithm',
        'extra:subject_organization_name',
        'extra:subject_organization_unit_name',
        'extra:subject_country',
        'extra:subject_state_or_province_name',
        'extra:subject_locality_name',
        'extra:subject_street_address',
        'extra:subject_postal_code',
        'extra:subject_surname',
        'extra:subject_given_name',
        'extra:subject_email_address',
        'extra:subject_business_category',
        'extra:subject_serial_number',
        'extra:issuer_organization_name',
        'extra:issuer_organization_unit_name',
        'extra:issuer_country',
        'extra:issuer_state_or_province_name',
        'extra:issuer_locality_name',
        'extra:issuer_street_address',
        'extra:issuer_postal_code',
        'extra:issuer_surname',
        'extra:issuer_given_name',
        'extra:issuer_email_address',
        'extra:issuer_business_category',
        'extra:issuer_serial_number',
        'extra:sha256_fingerprint',
        'extra:sha512_fingerprint',
        'extra:md5_fingerprint',
        'extra:cert_valid',
        'extra:self_signed',
        'extra:cert_expired',
        'extra:validation_level',
        'extra:auth_tls_response',
        'extra:auth_ssl_response',
        'extra:tlsv13_support',
        'extra:tlsv13_cipher',
    ]),
    ("csv_Accessible-RDP", [
        "time.source",
        "source.ip",
        ("source.port","port"),
        "source.reverse_dns",
        ("classification.identifier", "tag"),
        "extra:handshake",
        "source.asn",
        "source.geolocation.cc",
        "source.geolocation.region",
        "source.geolocation.city",
        "extra:rdp_protocol",
        "extra:cert_length", 
        "extra:subject_common_name",
        "extra:issuer_common_name",
        "extra:cert_issue_date",
        "extra:cert_expiration_date",
        "extra:sha1_fingerprint",
        'extra:cert_serial_number',
        'extra:ssl_version',
        'extra:signature_algorithm',
        'extra:key_algorithm',
        'extra:sha256_fingerprint',
        'extra:sha512_fingerprint',
        'extra:md5_fingerprint',
        'extra:naics',
        'extra:sic',
        'extra:sector',
        'extra:tlsv13_support',
        'extra:tlsv13_cipher',
        'extra:cve20190708_vulnerable',
        'extra:bluekeep_vulnerabe',
        ]),
    ("csv_Accessible-SMB", [
        # "timestamp","ip","port","hostname","asn","geo","region","city","naics","sic","smb_implant","arch","key"
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:smb_implant',
        'extra:arch',
        'extra:key',
    ]),
    ("csv_Accessible-Telnet", [
        #"timestamp","ip","protocol","port","hostname","tag","asn","geo","region","city","naics","sic","banner"
        'time.source',
        'source.ip',
        'protocol.transport',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:banner',
    ]),
    ("csv_Accessible-VNC", [
        #"timestamp","ip","port","hostname","asn","geo","region","city","naics","sic","product","banner"
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:product',
        'extra:banner',
    ]),
    ("csv_Block-Listed-IP", [
        "time.source",
        "source.ip",
        "source.reverse_dns",
        "extra:source",
        "extra:reason",
        "source.asn",
        "source.geolocation.cc",
        "source.geolocation.region",
        "source.geolocation.city",
        "extra:naics",
        "extra:sic", 
        "extra:sector", 
    ]),
    ("csv_Drone", [
        #"timestamp","ip","port","asn","geo","region","city","hostname","type","infection","url","agent","cc_ip","cc_port","cc_asn","cc_geo","cc_dns","count","proxy","application","p0f_genre","p0f_detail","machine_name","id","naics","sic","cc_naics","cc_sic","sector","cc_sector","ssl_cipher","family","tag","public_source",
        'time.source',
        'source.ip',
        'source.port',
        'source.asn', 
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'source.reverse_dns',
        ('protocol.transport','type'), #Will often fail and be put in extra:type instead because shadowserver puts "http" as value
        ('malware.name', 'infection'),
        'destination.url',
        'extra:user_agent', 
        ('destination.ip', 'cc_ip'),
        ('destination.port', 'cc_port'),
        ('destination.asn', 'cc_asn'),
        ('destination.geolocation.cc', 'cc_geo'),
        ('destination.fqdn', 'cc_dns'), 
        ExtraColumn('connection_count','count'),
        'extra:proxy',
        'protocol.application',
        ExtraColumn('os.name','p0f_genre'), 
        ExtraColumn('os.version','p0f_detail'),
        'extra:machine_name',
        'extra:id',
        'extra:naics',
        'extra:sic', 
        ExtraColumn('destination.naics', 'cc_naics'),
        ExtraColumn('destination.sic', 'cc_sic'),
        ExtraColumn('destination.sector', 'cc_sector'),
        'extra:sector',
        'extra:ssl_cipher',
        'extra:family',
        'extra:tag',
        'extra:public_source',
        'extra:type', # In case protocol.transport fails. The report will not be 100% shadowserver compatible
    ]),
    ("csv_Drone-Brute-Force", [
        #"timestamp","ip","port","asn","geo","region","city","hostname","dest_ip","dest_port","dest_asn","dest_geo","dest_dns","service","naics","sic","dest_naics","dest_sic","sector","dest_sector","public_source","start_time","end_time","client_version","username","password","payload_url","payload_md5"
        'time.source',
        'source.ip', 
        'source.port',
        ('source.asn', 'asn'),
        ('source.geolocation.cc', 'geo'),
        ('source.geolocation.region', 'region'),
        ('source.geolocation.city', 'city'),
        ('source.reverse_dns', 'hostname'),
        ('destination.ip', 'dest_ip'),
        ('destination.port', 'dest_port'),
        ('destination.asn', 'dest_asn'),
        ('destination.geolocation.cc', 'dest_geo'),
        ('destination.fqdn', 'dest_dns'),
        ('protocol.application', 'service'),
        'extra:naics',
        'extra:sic', 
        ExtraColumn('destination.naics', 'dest_naics'),
        ExtraColumn('destination.sic', 'dest_sic'),
        'extra:sector',
        ExtraColumn('destination.sector', 'dest_sector'),
        'extra:public_source',
        'extra:start_time',
        'extra:end_time',
        'extra:client_version',
        'destination.account',
        'extra:password',
        'extra:payload_url',
        'extra:payload_md5',
    ]),
    ("csv_Microsoft-Sinkhole", [
        #"timestamp","ip","asn","geo","url","type","http_agent","tor","src_port","p0f_genre","p0f_detail","hostname","dst_port","http_host","http_referer","http_referer_ip","http_referer_asn","http_referer_geo","dst_ip","dst_asn","dst_geo","naics","sic","http_referer_naics","http_referer_sic","sector","ssl_cipher","application","version"
        'time.source',
        'source.ip',
        'source.asn',
        'source.geolocation.cc',
        'destination.url',
        'malware.name',
        'extra:http_agent',
        'source.tor_node',
        ('source.port', 'src_port'),
        ExtraColumn('os.name','p0f_genre'), 
        ExtraColumn('os.version','p0f_detail'),
        'source.reverse_dns',
        'destination.port',
        'destination.fqdn',
        'extra:http_referer',
        'extra:http_referer_ip',
        'extra:http_referer_asn',
        'extra:http_referer_geo',
        'destination.ip',
        'destination.asn',
        'destination.geolocation.cc',
        'extra:naics',
        'extra:sic',
        'extra:http_referer_naics',
        'extra:http_referer_sic',
        'extra:sector',
        'extra:ssl_cipher',
        'extra:application',
        'extra:version',
    ]),
    ("csv_Open-Elasticsearch", [
        #"timestamp","ip","protocol","port","hostname","tag","version","asn","geo","region","city","naics","sic","ok","name","cluster_name","status","build_hash","build_timestamp","build_snapshot","lucene_version","tagline","sector"
        'time.source',
        'source.ip',
        'protocol.transport',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'extra:version',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:ok',
        'extra:name',
        'extra:cluster_name',
        'extra:status',
        'extra:build_hash',
        'extra:build_timestamp',
        'extra:build_snapshot',
        'extra:lucene_version',
        'extra:tagline',
        'extra:sector',
    ]),
    ("csv_Open-Memcached", [
        #"timestamp","ip","protocol","port","hostname","tag","version","asn","geo","region","city","naics","sic","pid","pointer_size","uptime","time","curr_connections","total_connections","sector"
        'time.source',
        'source.ip',
        'protocol.transport',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'extra:version',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:pid',
        'extra:pointer_size',
        'extra:uptime',
        'extra:time',
        'extra:curr_connections',
        'extra:total_connections',
        'extra:sector',
    ]),
    ("csv_Open-LDAP", [
        #"timestamp","ip","protocol","port","hostname","tag","asn","geo","region","city","naics","sic","size","configuration_naming_context","current_time","default_naming_context","dns_host_name","domain_controller_functionality","domain_functionality","ds_service_name","forest_functionality","highest_committed_usn","is_global_catalog_ready","is_synchronized","ldap_service_name","naming_contexts","root_domain_naming_context","schema_naming_context","server_name","subschema_subentry","supported_capabilities","supported_control","supported_ldap_policies","supported_ldap_version","supported_sasl_mechanisms"
        'time.source', 
        'source.ip',
        'protocol.transport', 
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:size',
        'extra:configuration_naming_context',
        'extra:current_time',
        'extra:default_naming_context',
        ('source.local_hostname', 'dns_host_name'),
        'extra:domain_controller_functionality',
        'extra:domain_functionality',
        'extra:ds_service_name',
        'extra:forest_functionality',
        'extra:highest_committed_usn',
        'extra:is_global_catalog_ready',
        'extra:is_synchronized',
        'extra:ldap_service_name',
        'extra:naming_contexts',
        'extra:root_domain_naming_context',
        'extra:schema_naming_context',
        'extra:server_name',
        'extra:subschema_subentry',
        'extra:supported_capabilities',
        'extra:supported_control',
        'extra:supported_ldap_policies',
        'extra:supported_ldap_version',
        'extra:supported_sasl_mechanisms',
    ]),
    ("csv_Open-LDAP-TCP", [
        #"timestamp","ip","protocol","port","hostname","tag","asn","geo","region","city","naics","sic","size","configuration_naming_context","current_time","default_naming_context","dns_host_name","domain_controller_functionality","domain_functionality","ds_service_name","forest_functionality","highest_committed_usn","is_global_catalog_ready","is_synchronized","ldap_service_name","naming_contexts","root_domain_naming_context","schema_naming_context","server_name","subschema_subentry","supported_capabilities","supported_control","supported_ldap_policies","supported_ldap_version","supported_sasl_mechanisms"
        'time.source', 
        'source.ip',
        'protocol.transport', 
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:size',
        'extra:configuration_naming_context',
        'extra:current_time',
        'extra:default_naming_context',
        ('source.local_hostname', 'dns_host_name'),
        'extra:domain_controller_functionality',
        'extra:domain_functionality',
        'extra:ds_service_name',
        'extra:forest_functionality',
        'extra:highest_committed_usn',
        'extra:is_global_catalog_ready',
        'extra:is_synchronized',
        'extra:ldap_service_name',
        'extra:naming_contexts',
        'extra:root_domain_naming_context',
        'extra:schema_naming_context',
        'extra:server_name',
        'extra:subschema_subentry',
        'extra:supported_capabilities',
        'extra:supported_control',
        'extra:supported_ldap_policies',
        'extra:supported_ldap_version',
        'extra:supported_sasl_mechanisms',
    ]),
    ("csv_Open-MongoDB", [
        #"timestamp","ip","protocol","port","hostname","tag","version","asn","geo","region","city","naics","sic","gitversion","sysinfo","opensslversion","allocator","javascriptengine","bits","maxbsonobjectsize","ok","visible_databases","sector"
        'time.source',
        'source.ip',
        'protocol.transport',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'extra:version',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:naics',
        'extra:sic',
        'extra:gitversion',
        'extra:sysinfo',
        'extra:opensslversion',
        'extra:allocator',
        'extra:javascriptengine',
        'extra:bits',
        'extra:maxbsonobjectsize',
        'extra:ok',
        'extra:visible_databases',
        'extra:sector',
    ]),
    ("csv_Sinkhole-HTTP-Drone", [
        #"timestamp","ip","asn","geo","url","type","http_agent","tor","src_port","p0f_genre","p0f_detail","hostname","dst_port","http_host","http_referer","http_referer_ip","http_referer_asn","http_referer_geo","dst_ip","dst_asn","dst_geo","naics","sic","http_referer_naics","http_referer_sic","sector","ssl_cipher","application","version"
        'time.source',
        'source.ip',
        'source.asn', 
        'source.geolocation.cc',
        'destination.url',
        ('malware.name', 'infection'),
        'extra:user_agent', 
        'source.tor_node',
        'source.port',
        ExtraColumn('os.name','p0f_genre'), 
        ExtraColumn('os.version','p0f_detail'),
        'source.reverse_dns',
        'destination.port',
        'destination.fqdn', 
        'extra:http_referer',
        'extra:http_referer_ip',
        'extra:http_referer_asn',
        'extra:http_referer_geo',
        'destination.ip',
        'destination.asn',
        'destination.geolocation.cc',
        'extra:naics',
        'extra:sic', 
        'extra:http_referer_naics',
        'extra:http_referer_sic',
        'extra:sector',
        'extra:ssl_cipher',
        'extra:application',
        'extra:version',
    ]),
    ("csv_SSL-POODLE-Vulnerable-Servers", [
        #"timestamp","ip","port","hostname","tag","handshake","asn","geo","region","city","cipher_suite","ssl_poodle","cert_length","subject_common_name","issuer_common_name","cert_issue_date","cert_expiration_date","sha1_fingerprint","cert_serial_number","ssl_version","signature_algorithm","key_algorithm","subject_organization_name","subject_organization_unit_name","subject_country","subject_state_or_province_name","subject_locality_name","subject_street_address","subject_postal_code","subject_surname","subject_given_name","subject_email_address","subject_business_category","subject_serial_number","issuer_organization_name","issuer_organization_unit_name","issuer_country","issuer_state_or_province_name","issuer_locality_name","issuer_street_address","issuer_postal_code","issuer_surname","issuer_given_name","issuer_email_address","issuer_business_category","issuer_serial_number","naics","sic","sector","sha256_fingerprint","sha512_fingerprint","md5_fingerprint","http_response_type","http_code","http_reason","content_type","http_connection","www_authenticate","set_cookie","server_type","content_length","transfer_encoding","http_date","cert_valid","self_signed","cert_expired","browser_trusted","validation_level","browser_error","tlsv13_support","tlsv13_cipher","raw_cert","raw_cert_chain"
        'time.source', 
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'classification.identifier',
        'extra:handshake',
        'source.asn',
        'source.geolocation.cc',
        'source.geolocation.region',
        'source.geolocation.city',
        'extra:cipher_suite',
        'extra:ssl_poodle',
        'extra:cert_length',
        'extra:subject_common_name',
        'extra:issuer_common_name',
        'extra:cert_issue_date',
        'extra:cert_expiration_date',
        'extra:sha1_fingerprint',
        'extra:cert_serial_number',
        'extra:ssl_version',
        'extra:signature_algorithm',
        'extra:key_algorithm',
        'extra:subject_organization_name',
        'extra:subject_organization_unit_name',
        'extra:subject_country',
        'extra:subject_state_or_province_name',
        'extra:subject_locality_name',
        'extra:subject_street_address',
        'extra:subject_postal_code',
        'extra:subject_surname',
        'extra:subject_given_name',
        'extra:subject_email_address',
        'extra:subject_business_category',
        'extra:subject_serial_number',
        'extra:issuer_organization_name',
        'extra:issuer_organization_unit_name',
        'extra:issuer_country',
        'extra:issuer_state_or_province_name',
        'extra:issuer_locality_name',
        'extra:issuer_street_address',
        'extra:issuer_postal_code',
        'extra:issuer_surname',
        'extra:issuer_given_name',
        'extra:issuer_email_address',
        'extra:issuer_business_category',
        'extra:issuer_serial_number',
        'extra:naics',
        'extra:sic',
        'extra:sector',
        'extra:sha256_fingerprint',
        'extra:sha512_fingerprint',
        'extra:md5_fingerprint',
        'extra:http_response_type',
        'extra:http_code',
        'extra:http_reason',
        'extra:content_type',
        'extra:http_connection',
        'extra:www_authenticate',
        'extra:set_cookie',
        'extra:server_type',
        'extra:content_length',
        'extra:transfer_encoding',
        'extra:http_date',
        'extra:cert_valid',
        'extra:self_signed',
        'extra:cert_expired',
        'extra:browser_trusted',
        'extra:validation_level',
        'extra:browser_error',
        'extra:tlsv13_support',
        'extra:tlsv13_cipher',
    ]),
])

# Used to create ${inline_events}
inline_formats = table_formats_with_default_titles([
    ("csv_Amplification-DDoS-Victim", [
         "time.source",
         ('source.ip' , 'victim_ip'),
         ('source.port' , 'victim_target_port'),
         ('source.reverse_dns' , 'victim_hostname'),
         'extra:end_time',
         ('destination.port' , 'ddos_source_port'),
    ]),
    ("csv_Accessible-FTP", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
    ]),
    ("csv_Accessible-RDP", [
        "time.source",
        "source.ip",
        ("source.port", "port"),
        'source.reverse_dns',
        'extra:subject_common_name',
        'extra:cve20190708_vulnerable',
        'extra:bluekeep_vulnerable',
    ]),
    ("csv_Accessible-SMB", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'extra:smb_implant',
    ]),
    ("csv_Accessible-Telnet", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
    ]),
    ("csv_Accessible-VNC", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
    ]),
    ("csv_Block-Listed-IP", [
        "time.source",
        "source.ip",
        "source.reverse_dns",
        "extra:source",
        "extra:reason",
    ]),    
    ("csv_Drone", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        ('classification.identifier', 'malware_family'),
        'destination.url',
        ('destination.ip', 'cc_ip'),
        ('destination.port', 'cc_port'),
        ('destination.fqdn', 'cc_dns'), 
        'extra:public_source',
    ]),
    ("csv_Drone-Brute-Force", [
        'time.source',
        'source.ip', 
        'source.port',
        ('source.reverse_dns', 'hostname'),
        ('destination.ip', 'dest_ip'),
        ('destination.port', 'dest_port'),
        'extra:start_time',
        'extra:end_time',
        'destination.account',
        'extra:password',
    ]),
    ("csv_Microsoft-Sinkhole", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        ('classification.identifier', 'malware_family'),
        'destination.url',
        'destination.ip',
        'destination.port',
    ]),
    ("csv_Open-Elasticsearch", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'extra:version',
        'extra:name',
    ]),
    ("csv_Open-LDAP", [
        'time.source', 
        'source.ip',
        'source.port',
        'protocol.transport', 
        'source.reverse_dns',
    ]),
    ("csv_Open-LDAP-TCP", [
        'time.source', 
        'source.ip',
        'source.port',
        'protocol.transport', 
        'source.reverse_dns',
    ]),
    ("csv_Open-Memcached", [
        'time.source',
        'source.ip',
        'source.port',
        'protocol.transport',
        'source.reverse_dns',
        'extra:version',
    ]),
    ("csv_Open-MongoDB", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'extra:version',
        'extra:visible_databases',
        'extra:opensslversion',
    ]),
    ("csv_Sinkhole-HTTP-Drone", [
        'time.source',
        'source.ip',
        'source.port',
        'source.reverse_dns',
        ('classification.identifier', 'malware_family'),
        'destination.url',
        'destination.ip',
        'destination.port',
    ]),
    ("csv_DNS-open-resolvers", [
        "source.asn",
        "source.ip",
        "time.source",
        ]),
    ("csv_Open-Portmapper", [
        "source.asn",
        "source.ip",
        "time.source",
        ]),
    ("csv_Open-SNMP", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:system_desc",
        ]),
    ("csv_Open-MSSQL", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:mssql_version",
        "source.local_hostname",
        ExtraColumn("instance_name", "instance_name"),
        ]),
    ("csv_Open-Chargen", [
        "source.asn",
        "source.ip",
        "time.source",
        ]),
    ("csv_Open-IPMI", [
        "source.asn",
        "source.ip",
        "time.source",
        ]),
    ("csv_Open-NetBIOS", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:workgroup_name",
        "extra:machine_name",
        ]),
    ("csv_NTP-Monitor", [
        "source.asn",
        "source.ip",
        "time.source",
        ]),
    ("csv_Open-mDNS", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:workstation_info",
        ]),
    ("csv_Open-Redis", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:redis_version",
        ]),
    ("csv_Open-SSDP", [
        "source.asn",
        "source.ip",
        "time.source",
        "extra:ssdp_server",
        ]),
    ("csv_Ssl-Freak-Scan", [
        "source.asn",
        "source.ip",
        "time.source",
        "source.reverse_dns",
        "extra:subject_common_name",
        "extra:issuer_common_name",
        "extra:freak_cipher_suite",
        ]),
    ("csv_SSL-POODLE-Vulnerable-Servers", [
        #"timestamp","ip","port","hostname","tag","handshake","asn","geo","region","city","cipher_suite","ssl_poodle","cert_length","subject_common_name","issuer_common_name","cert_issue_date","cert_expiration_date","sha1_fingerprint","cert_serial_number","ssl_version","signature_algorithm","key_algorithm","subject_organization_name","subject_organization_unit_name","subject_country","subject_state_or_province_name","subject_locality_name","subject_street_address","subject_postal_code","subject_surname","subject_given_name","subject_email_address","subject_business_category","subject_serial_number","issuer_organization_name","issuer_organization_unit_name","issuer_country","issuer_state_or_province_name","issuer_locality_name","issuer_street_address","issuer_postal_code","issuer_surname","issuer_given_name","issuer_email_address","issuer_business_category","issuer_serial_number","naics","sic","sector","sha256_fingerprint","sha512_fingerprint","md5_fingerprint","http_response_type","http_code","http_reason","content_type","http_connection","www_authenticate","set_cookie","server_type","content_length","transfer_encoding","http_date","cert_valid","self_signed","cert_expired","browser_trusted","validation_level","browser_error","tlsv13_support","tlsv13_cipher","raw_cert","raw_cert_chain"
        'time.source', 
        'source.ip',
        'source.port',
        'source.reverse_dns',
        'extra:handshake',
        'extra:cipher_suite',
        'extra:cert_length',
        'extra:subject_common_name',
        'extra:cert_issue_date',
        'extra:cert_expiration_date',
    ]),
    ("csv_Ssl-Scan", [
        "source.asn",
        "source.ip",
        "time.source",
        "source.reverse_dns",
        "extra:subject_common_name",
        "extra:issuer_common_name",
        ]),
    ])

# Minimum age of the newest of a group of directives being aggregated
#
# The value should be chosen such that it's very unlikely that any more
# directives will be added to the event database that would end up in
# the same aggregation when the newest directive has reached at least
# this age.
#minimum_directive_age = datetime.timedelta(minutes=15)
minimum_directive_age = datetime.timedelta(minutes=1)

#minimum_observation_age = datetime.timedelta(hours=2)
minimum_observation_age = datetime.timedelta(minutes=1)


def create_notifications(context):
    """

    Args:
        context:

    Returns:

    """

    if context.directive.notification_format == "shadowserver":
        if context.directive.event_data_format not in inline_formats and context.directive.event_data_format not in csv_formats:
            # We don't handle this event type
            return None

        template_file = full_template_filename(context.config["template_dir"],context.directive.template_name)
        if not os.path.isfile(template_file): 
            raise RuntimeError("Template file %s missing" % (template_file,))

        context.logger.debug("10shadowservercsv.py: notification format is shadowserver and we have something to do")
        format_spec = None
        format_spec2 = None
        if context.directive.event_data_format in inline_formats: 
            format_spec2 = inline_formats.get(context.directive.event_data_format)
            events = context.load_events(format_spec2.event_table_columns())
            context.logger.debug("Events loaded: %r",events)
        if context.directive.event_data_format in csv_formats: 
            format_spec = csv_formats.get(context.directive.event_data_format)
            context.logger.debug("format_spec: %r",format_spec)
            context.logger.debug("Events loaded: %r", context.load_events(format_spec.event_table_columns()))

        # Copy substitutions from the context.
        # This way we can edit the variables in this script
        # without changing the context.
        substitution_variables = copy.copy(context.substitutions)
        context.logger.debug("Format_spec2: %r",format_spec2)
        if format_spec2 is not None: 
#            context.logger.debug("Format_spec2: column_keys: %r(%r)",type(format_spec2.column_keys()),format_spec2.column_keys())
#            context.logger.debug("Format_spec2: column_titles: %r(%r)",type(format_spec2.column_titles()),format_spec2.column_titles())
#            context.logger.debug("Format_spec2: event_table_columns: %r(%r)",type(format_spec2.event_table_columns()),format_spec2.event_table_columns())
            substitution_variables["inline_events"] = format_as_csv(format_spec2,events)
        else:
            substitution_variables["inline_events"] = "Inga inline data att visa. Se bifogad CSV-fil istället"

        if format_spec is not None:
#            if (context.age_of_newest_directive() < minimum_directive_age
#                or context.age_of_observation() < minimum_observation_age):
#            if (context.age_of_newest_directive() < minimum_directive_age):
#                return Postponed

            return context.mail_format_as_csv(format_spec, substitutions=substitution_variables,attach_event_data=True)
        elif format_spec2 is not None:
            return context.mail_format_as_csv(format_spec2, substitutions=substitution_variables)
        else:
            # We should never be here
            raise RuntimeError("Strange. Couldn't format events according to directives")

    return None
